/*
 *        spAudio_Plugin.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spAudioP.h>
#include <sp/spInputPluginP.h>
#include <sp/spRwPluginP.h>

#include "spAudio_Plugin.h"

#define SP_MAX_NUM_AUDIO_PLUGIN 16

static int sp_num_audio_plugin = -1;
static char sp_audio_plugin_list[SP_MAX_NUM_AUDIO_PLUGIN][SP_MAX_PATHNAME];

int spGetNumAudioDriverPluginFromHostData(spPluginHostData host_data)
{
    spDebug(1, "spGetNumAudioDriverPlugin", "in: sp_num_audio_plugin = %d\n", sp_num_audio_plugin);
    
    if (sp_num_audio_plugin < 0) {
        char buf[SP_MAX_PATHNAME];
        
        sp_num_audio_plugin = 0;

        while (spSearchPluginDeviceFromHostData(host_data, sp_num_audio_plugin, SP_PLUGIN_RW,
                                                SP_PLUGIN_DEVICE_AUDIO, buf) == SP_TRUE) {
            strcpy(sp_audio_plugin_list[sp_num_audio_plugin], spGetBaseName(buf));
            spRemoveSuffix(sp_audio_plugin_list[sp_num_audio_plugin], NULL);
            
            spDebug(1, "spGetNumAudioDriverPlugin", "sp_audio_plugin_list[%d] = %s\n",
                    sp_num_audio_plugin, sp_audio_plugin_list[sp_num_audio_plugin]);
            sp_num_audio_plugin++;

            if (sp_num_audio_plugin >= SP_MAX_NUM_AUDIO_PLUGIN) {
                break;
            }
        }

    }
    
    spDebug(1, "spGetNumAudioDriverPlugin", "done: sp_num_audio_plugin = %d\n", sp_num_audio_plugin);
    
    return sp_num_audio_plugin;
}

int spGetNumAudioDriverPlugin(void)
{
    return spGetNumAudioDriverPluginFromHostData(spGetPluginHostData());
}

const char *spGetAudioDriverNamePluginFromHostData(spPluginHostData host_data, int index)
{
    if (sp_num_audio_plugin < 0) {
        spGetNumAudioDriverPluginFromHostData(host_data);
    }

    if (sp_num_audio_plugin > 0
        && index >= 0 && index < sp_num_audio_plugin) {
        return sp_audio_plugin_list[index];
    } else {
        return NULL;
    }
}

const char *spGetAudioDriverNamePlugin(int index)
{
    return spGetAudioDriverNamePluginFromHostData(spGetPluginHostData(), index);
}

spBool spInitAudioDriverPlugin(spAudio audio, const char *driver_name, void *data)
{
    spPluginHostData host_data = (spPluginHostData)data;

    if (host_data == NULL) {
        spDebug(10, "spInitAudioDriverPlugin", "host_data is null\n");
        host_data = spGetPluginHostData();
    }
    
    if (strnone(driver_name)) {
        spDebug(10, "spInitAudioDriverPlugin", "driver_name is null\n");
        return SP_FALSE;
    }

    spDebug(10, "spInitAudioDriverPlugin", "driver_name = %s, call spLoadPluginFromHostData\n", driver_name);
    
    if ((audio->plugin = spLoadPluginFromHostData(host_data, driver_name)) == NULL) {
        spDebug(10, "spInitAudioDriverPlugin", "spLoadPluginFromHostData failed\n");
        return SP_FALSE;
    }
    if (spIsAudioDriverPlugin(audio->plugin) == SP_FALSE) {
        spDebug(10, "spInitAudioDriverPlugin", "spIsAudioDriverPlugin == SP_FALSE\n");
        spClosePlugin(audio->plugin);
        audio->plugin = NULL;
        return SP_FALSE;
    }
    
    spDebug(10, "spInitAudioDriverPlugin", "done\n");
    
    return SP_TRUE;
}

spBool spGetNumAudioDevicePlugin(spAudio audio, int *num_device)
{
    return spGetPluginNumDevice(audio->plugin, num_device);
}

const char *spGetAudioDeviceNamePlugin(spAudio audio, int device_index)
{
    return spGetPluginDeviceName(audio->plugin, device_index);
}

spBool spSelectAudioDevicePlugin(spAudio audio, int device_index)
{
    return spSelectPluginDevice(audio->plugin, device_index);
}

unsigned long spGetAudioCapsPlugin(spAudio audio)
{
    return spGetPluginCaps(audio->plugin);
}

spBool spDisplayAudioOptionDialogPlugin(spAudio audio)
{
    return spDisplayPluginCustomOptionDialog(audio->plugin);
}

spBool spOpenAudioDevicePlugin(spAudio audio, const char *mode)
{
    if (spOpenPlugin(audio->plugin, NULL, mode) == SP_PLUGIN_ERROR_SUCCESS) {
        spGetPluginSampleBit(audio->plugin, &audio->samp_bit);
        spGetPluginSampleRate(audio->plugin, &audio->samp_rate);
        spGetPluginBufferSize(audio->plugin, &audio->buffer_size);
        spGetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_NUM_BUFFER, (char *)&audio->num_buffer);
        spGetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_BLOCKING_MODE, (char *)&audio->block_mode);
        
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spCloseAudioDevicePlugin(spAudio audio)
{
    return spClosePlugin(audio->plugin);
}

spBool spSetAudioSampleBitPlugin(spAudio audio)
{
    return spSetPluginSampleBit(audio->plugin, audio->samp_bit);
}

spBool spSetAudioSampleRatePlugin(spAudio audio)
{
    return spSetPluginSampleRate(audio->plugin, audio->samp_rate);
}

spBool spSetAudioChannelPlugin(spAudio audio)
{
    return spSetPluginChannel(audio->plugin, audio->num_channel);
}

spBool spSetAudioBufferSizePlugin(spAudio audio)
{
    return spSetPluginBufferSize(audio->plugin, audio->buffer_size);
}

spBool spSetAudioNumBufferPlugin(spAudio audio)
{
    return spSetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_NUM_BUFFER, (char *)&audio->num_buffer);
}

spBool spSetAudioBlockingModePlugin(spAudio audio)
{
    return spSetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_BLOCKING_MODE, (char *)&audio->block_mode);
}

spAudioCallbackType spSetAudioCallbackFuncPlugin(spAudio audio)
{
    spAudioCallbackType supported_type = SP_AUDIO_NO_CALLBACK;

    if (spSetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_CALLBACK_FUNC, (char *)audio->call_func) == SP_TRUE) {
        spSetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_CALLBACK_HOST_DATA, (char *)audio);
        spSetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_CALLBACK_DATA, (char *)audio->call_data);
        
        spSetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_CALLBACK_TYPE, (char *)&audio->call_type);
        spGetPluginOtherInfo(audio->plugin, SP_PLUGIN_INFO_CALLBACK_TYPE, (char *)&supported_type);
        
        return supported_type;
    } else {
        return SP_AUDIO_NO_CALLBACK;
    }
}

long spReadAudioPlugin(spAudio audio, char *data, long length)
{
    return spReadPlugin(audio->plugin, data, length);
}

long spWriteAudioPlugin(spAudio audio, char *data, long length)
{
    return spWritePlugin(audio->plugin, data, length);
}

spBool spGetAudioOutputPositionPlugin(spAudio audio, spLong *position)
{
    if ((*position = spGetPluginCurrentPosition(audio->plugin)) >= 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spStopAudioPlugin(spAudio audio)
{
    return spStopPlugin(audio->plugin);
}

spBool spSyncAudioPlugin(spAudio audio)
{
    return spFlushPlugin(audio->plugin);
}

spBool spFreeAudioDriverPlugin(spAudio audio)
{
    spBool flag;

    if (audio->plugin == NULL) return SP_FALSE;
    
    spDebug(20, "spFreeAudioDriverPlugin", "in\n");
    
    flag = spFreePlugin(audio->plugin);
    audio->plugin = NULL;

    spDebug(20, "spFreeAudioDriverPlugin", "done: flag = %d\n", flag);
    
    return flag;
}

void spTerminateAudioPlugin(void)
{
    return;
}
