/*
 *        spWave_Wav.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spKanji.h>

#include <sp/spWave_Wav.h>

long spReadRiffFourCC(char *fourcc, DWORD *size, FILE *fp)
{
    long read_size;
    
    if (fread(fourcc, 1, 4, fp) != 4) {
        return 0;
    }
    read_size = 4;

    spFReadDWORD(size, 1, fp);
    read_size += 4;

    spDebug(100, "spReadRiffFourCC", "fourcc = %c%c%c%c, size = %ld, read_size = %ld\n",
            fourcc[0], fourcc[1], fourcc[2], fourcc[3], *size, read_size);
    
    return read_size;
}

const char *spGetWavFormatLabelFromTag(WORD wFormatTag)
{
    switch (wFormatTag) {
      case WAVE_FORMAT_UNKNOWN:
        return WAVE_FORMAT_UNKNOWN_STRING;
      case WAVE_FORMAT_PCM:
        return WAVE_FORMAT_PCM_STRING;
      case WAVE_FORMAT_ADPCM:
        return WAVE_FORMAT_ADPCM_STRING;
      case WAVE_FORMAT_IEEE_FLOAT:
        return WAVE_FORMAT_IEEE_FLOAT_STRING;
      case WAVE_FORMAT_ALAW:
        return WAVE_FORMAT_ALAW_STRING;
      case WAVE_FORMAT_MULAW:
        return WAVE_FORMAT_MULAW_STRING;
      case WAVE_FORMAT_OKI_ADPCM:
        return WAVE_FORMAT_OKI_ADPCM_STRING;
      case WAVE_FORMAT_DIGISTD:
        return WAVE_FORMAT_DIGISTD_STRING;
      case WAVE_FORMAT_DIGIFIX:
        return WAVE_FORMAT_DIGIFIX_STRING;
      default:
        return "Unknown";
    }
}

const char *spGetWavFormatLabel(WAVEHEADER wave_header)
{
    return spGetWavFormatLabelFromTag(wave_header.wFormatTag);
}

WORD spGetWavFormatTag(const char *format)
{
    if (strnone(format)
        || strcaseeq(format, WAVE_FORMAT_PCM_STRING)) {
        return WAVE_FORMAT_PCM;
    } else if (strcaseeq(format, WAVE_FORMAT_UNKNOWN_STRING)) {
        return WAVE_FORMAT_UNKNOWN;
    } else if (strcaseeq(format, WAVE_FORMAT_ADPCM_STRING)) {
        return WAVE_FORMAT_ADPCM;
    } else if (strcaseeq(format, WAVE_FORMAT_IEEE_FLOAT_STRING)) {
        return WAVE_FORMAT_IEEE_FLOAT;
    } else if (strcaseeq(format, WAVE_FORMAT_ALAW_STRING)) {
        return WAVE_FORMAT_ALAW;
    } else if (strcaseeq(format, WAVE_FORMAT_MULAW_STRING)) {
        return WAVE_FORMAT_MULAW;
    } else if (strcaseeq(format, WAVE_FORMAT_OKI_ADPCM_STRING)) {
        return WAVE_FORMAT_OKI_ADPCM;
    } else if (strcaseeq(format, WAVE_FORMAT_DIGISTD_STRING)) {
        return WAVE_FORMAT_DIGISTD;
    } else if (strcaseeq(format, WAVE_FORMAT_DIGIFIX_STRING)) {
        return WAVE_FORMAT_DIGIFIX;
    }
    
    return WAVE_FORMAT_UNKNOWN;
}

static long readWaveFormatInfo(WORD *pwFormatTag, WORD *pnChannels, DWORD *pnSamplesPerSec,
                               DWORD *pnAvgBytesPerSec, WORD *pnBlockAlign, WORD *pwBitsPerSample,
                               FILE *fp, spBool ex_flag, spBool check_format)
{
    long read_size;

    if (spFReadWORD(pwFormatTag, 1, fp) <= 0) {
        spDebug(1, "readWaveFormatInfo", "wFormatTag not found\n");
        return 0;
    }
    spDebug(1, NULL, "wFormatTag = %d, BYTE_ORDER = %d, SP_WAV_NEED_SWAP = %d\n",
            *pwFormatTag, BYTE_ORDER, SP_WAV_NEED_SWAP);
    
    read_size = 2;
    
    if (check_format == SP_TRUE) {
        switch (*pwFormatTag) {
          case WAVE_FORMAT_PCM:
          case WAVE_FORMAT_ALAW:
          case WAVE_FORMAT_MULAW:
            /*case WAVE_FORMAT_IEEE_FLOAT:*/
            break;
          default:
            spDebug(1, NULL, "%s format is not supported.\n", spGetWavFormatLabelFromTag(*pwFormatTag));
            return 0;
        }
    }
    
    if (spFReadWORD(pnChannels, 1, fp) <= 0) {
        spDebug(50, "readWaveFormatInfo", "read nChannels failed\n");
        return 0;
    }
    read_size += 2;
    
    if (spFReadDWORD(pnSamplesPerSec, 1, fp) <= 0) {
        spDebug(50, "readWaveFormatInfo", "read nSamplesPerSec failed\n");
        return 0;
    }
    read_size += 4;
    
    if (spFReadDWORD(pnAvgBytesPerSec, 1, fp) <= 0) {
        spDebug(50, "readWaveFormatInfo", "read nAvgBytesPerSec failed\n");
        return 0;
    }
    read_size += 4;
    
    if (spFReadWORD(pnBlockAlign, 1, fp) <= 0) {
        spDebug(50, "readWaveFormatInfo", "read nBlockAlign failed\n");
        return 0;
    }
    read_size += 2;

    if (ex_flag == SP_TRUE) {
        if (spFReadWORD(pwBitsPerSample, 1, fp) <= 0) {
            spDebug(50, "readWaveFormatInfo", "read wBitsPerSample failed\n");
            return 0;
        }
        read_size += 2;
    } else {
        *pwBitsPerSample = (*pnBlockAlign / *pnChannels) * 8;
    }
    
    spDebug(50, "readWaveFormatInfo", "done: read_size = %ld\n", read_size);
    
    return read_size;
}

static long readWavHeader(WAVEHEADER *wave_header, FILE *fp, spBool check_format, spBool after_riff)
{
    long read_size;
    long head_size = 0;
    long len = 0;
    DWORD ulSize;

    if (after_riff == SP_TRUE) {
        head_size = 8;
    } else {
        /* seek start of the file */
        spSeekFile(fp, 0L, SEEK_SET);

        if ((head_size = spReadRiffFourCC(wave_header->RiffID, &ulSize, fp)) <= 0
            || !strneq("RIFF", wave_header->RiffID, 4)) {
            spDebug(1, NULL, "Can't find RIFF chunk.\n");
            return 0;
        }
        wave_header->RiffSize = (DWORD)ulSize;
    }
    spDebug(50, "readWavHeader", "RIFF found\n");
    
    if (fread(wave_header->WaveID, 1, 4, fp) != 4 ||
        !strneq(wave_header->WaveID, "WAVE", 4)) {
        spDebug(1, NULL, "Can't find WAVE chunk.\n");
        return 0;
    }
    head_size += 4;
    spDebug(50, "readWavHeader", "WAVE found\n");

    for (;;) {
        if (fread(wave_header->FmtID, 1, 4, fp) != 4) {
            spDebug(1, NULL, "Can't find fmt chunk\n");
            return 0;
        }
        head_size += 4;
        
        spFReadDWORD(&wave_header->FmtSize, 1, fp);
        len = wave_header->FmtSize;
        head_size += 4;
        
        if (strneq("fmt ", wave_header->FmtID, 4))
            break;

        while (len > 0 && !feof(fp)) {
            getc(fp);
            len--;
            head_size += 1;
        }
    }
    
    if (len < 16) {
        spDebug(1, NULL, "Fmt chunk is too short.\n");
        return 0;
    }

    if ((read_size = readWaveFormatInfo(&wave_header->wFormatTag, &wave_header->nChannels,
                                        &wave_header->nSamplesPerSec, &wave_header->nAvgBytesPerSec,
                                        &wave_header->nBlockAlign, &wave_header->wBitsPerSample,
                                        fp, SP_TRUE, check_format)) <= 0) {
        spDebug(10, "readWavHeader", "readWaveFormatInfo failed\n");
        return 0;
    }
    head_size += read_size;
    len -= read_size;

    while (len > 0 && !feof(fp)) {
        getc(fp);
        len--;
        head_size += 1;
    }
    
    for (;;) {
        if (fread(wave_header->DataID, 1, 4, fp) != 4) {
            spDebug(1, NULL, "Sorry, missing data chunk\n");
            return 0;
        }
        head_size += 4;

        spFReadDWORD(&wave_header->nDataBytes, 1, fp);
        len = wave_header->nDataBytes;
        head_size += 4;
        
        if (strneq("data", wave_header->DataID, 4))
            break;

        while (len > 0 && !feof(fp)) {
            getc(fp);
            len--;
            head_size += 1;
        }
    }

    spDebug(10, NULL, "%s format, %d channel, %ld samp/sec\n",
            spGetWavFormatLabelFromTag(wave_header->wFormatTag), wave_header->nChannels,
            wave_header->nSamplesPerSec);
    spDebug(10, NULL, "%ld byte/sec, %d block align, %d bits/samp, %u data bytes\n",
            wave_header->nAvgBytesPerSec, wave_header->nBlockAlign,
            wave_header->wBitsPerSample, wave_header->nDataBytes);
    spDebug(10, NULL, "header size = %ld\n", head_size);

    return head_size;
}

long spReadWavHeader(WAVEHEADER *wave_header, FILE *fp)
{
    return readWavHeader(wave_header, fp, SP_TRUE, SP_FALSE);
}

long spParseWavHeader(WAVEHEADER *wave_header, FILE *fp)
{
    return readWavHeader(wave_header, fp, SP_FALSE, SP_FALSE);
}

long spWriteWavHeader(WAVEHEADER wave_header, long length, FILE *fp)
{
    long head_size = 0;

    switch (wave_header.wFormatTag) {
      case WAVE_FORMAT_PCM:
      case WAVE_FORMAT_ALAW:
      case WAVE_FORMAT_MULAW:
        /*case WAVE_FORMAT_IEEE_FLOAT:*/
        break;
      default:
        spDebug(1, NULL, "%s format is not supported.\n", spGetWavFormatLabelFromTag(wave_header.wFormatTag));
        return 0;
    }
    if (wave_header.nSamplesPerSec <= 0) {
        spDebug(1, NULL, "Invalid sampling frequency: %ld\n", (long)wave_header.nSamplesPerSec);
        return 0;
    }
    if (wave_header.nChannels <= 0) {
        spDebug(1, NULL, "Invalid number of channels: %d\n", wave_header.nChannels);
        return 0;
    }
    if (wave_header.wBitsPerSample < 8) {
        spDebug(1, NULL, "Invalid bits/sample: %d\n", wave_header.wBitsPerSample);
        return 0;
    }
    
    /* seek start of the file */
    spSeekFile(fp, 0L, SEEK_SET);
    
    /* compute parameters */
    wave_header.FmtSize = 16;
    wave_header.nBlockAlign = wave_header.nChannels * (wave_header.wBitsPerSample / 8);
    wave_header.nAvgBytesPerSec = wave_header.nSamplesPerSec * wave_header.nBlockAlign;
    wave_header.nDataBytes = length * wave_header.nChannels * (wave_header.wBitsPerSample / 8);
    wave_header.RiffSize = wave_header.nDataBytes + 36;
    
    fputs("RIFF", fp);
    head_size += 4;
    spFWriteDWORD(&wave_header.RiffSize, 1, fp);
    head_size += 4;
    fputs("WAVE", fp);
    head_size += 4;
    fputs("fmt ", fp);
    head_size += 4;
    spFWriteDWORD(&wave_header.FmtSize, 1, fp);
    head_size += 4;
    
    spFWriteWORD(&wave_header.wFormatTag, 1, fp);
    head_size += 2;
    spFWriteWORD(&wave_header.nChannels, 1, fp);
    head_size += 2;
    spFWriteDWORD(&wave_header.nSamplesPerSec, 1, fp);
    head_size += 4;
    spFWriteDWORD(&wave_header.nAvgBytesPerSec, 1, fp);
    head_size += 4;
    spFWriteWORD(&wave_header.nBlockAlign, 1, fp);
    head_size += 2;
    spFWriteWORD(&wave_header.wBitsPerSample, 1, fp);
    head_size += 2;
    
    fputs("data", fp);
    head_size += 4;
    spFWriteDWORD(&wave_header.nDataBytes, 1, fp);
    head_size += 4;

    spDebug(10, NULL, "%s format, %d channel, %d samp/sec\n",
            spGetWavFormatLabelFromTag(wave_header.wFormatTag), wave_header.nChannels,
            wave_header.nSamplesPerSec);
    spDebug(10, NULL, "%d byte/sec, %d block align, %d bits/samp, %u data bytes\n",
            wave_header.nAvgBytesPerSec, wave_header.nBlockAlign,
            wave_header.wBitsPerSample, wave_header.nDataBytes);
    spDebug(10, NULL, "header size = %ld\n", head_size);

    return head_size;
}

spBool spIsWavFile(const char *filename)
{
    FILE *fp;
    WAVEHEADER wave_header;
    spBool flag = SP_FALSE;
    
    if (filename != NULL && (fp = spOpenFile(filename, "rb")) != NULL) {
        if (spReadWavHeader(&wave_header, fp) > 0) {
            flag = SP_TRUE;
        }
        spCloseFile(fp);
        spDebug(20, "spIsWavFile", "%s: flag = %d\n", filename, flag);
    } else {
        spDebug(10, "spIsWavFile", "open error: %s\n", filename);
    }

    return flag;
}

static void updateWavInfo(WAVEHEADER *wave_header, spWaveInfo *wave_info)
{
    strcpy(wave_info->file_type, SP_WAVE_WAV_ID);
    strcpy(wave_info->file_desc, spGetWavFormatLabel(*wave_header));
    strcpy(wave_info->file_filter, "*.wav");

    return;
}

spBool spReadWavInfo(spWaveInfo *wave_info, FILE *fp)
{
    WAVEHEADER wave_header;
    long header_size;

    if (wave_info == NULL || fp == NULL) return SP_FALSE;

    if ((header_size = spReadWavHeader(&wave_header, fp)) <= 0) {
        return SP_FALSE;
    }

    updateWavInfo(&wave_header, wave_info);
    
    wave_info->header_size = header_size;
    wave_info->num_channel = wave_header.nChannels;
    wave_info->samp_bit = wave_header.wBitsPerSample;
    wave_info->samp_rate = (double)wave_header.nSamplesPerSec;
    wave_info->bit_rate = wave_header.nAvgBytesPerSec * 8;
    wave_info->length = wave_header.nDataBytes
        / (wave_header.nChannels * wave_header.wBitsPerSample / 8);
    
    return SP_TRUE;
}

spBool spWriteWavInfo(spWaveInfo *wave_info, FILE *fp)
{
    WAVEHEADER wave_header;
    long header_size;

    if (wave_info == NULL || fp == NULL) return SP_FALSE;

    /*wave_header.wFormatTag = WAVE_FORMAT_PCM;*/
    wave_header.wFormatTag = spGetWavFormatTag(wave_info->file_desc);
    wave_header.nChannels = wave_info->num_channel;
    wave_header.wBitsPerSample = wave_info->samp_bit;
    wave_header.nSamplesPerSec = (DWORD)spRound(wave_info->samp_rate);
    
    if ((header_size = spWriteWavHeader(wave_header, (long)wave_info->length, fp)) <= 0) {
        return SP_FALSE;
    }
    updateWavInfo(&wave_header, wave_info);

    wave_info->header_size = header_size;

    return SP_TRUE;
}

long _spReadWavData(spWaveInfo *wave_info, char *data, long length, FILE *fp)
{
    long len;
    
    if (wave_info == NULL || fp == NULL) return -1;

    if (wave_info->samp_bit == 16) {
        len = freadshort((short *)data, length, SP_WAV_NEED_SWAP, fp);
    } else if (wave_info->samp_bit == 64
               && strcaseeq(wave_info->file_desc, WAVE_FORMAT_IEEE_FLOAT_STRING)) {
        len = freaddouble((double *)data, length, SP_WAV_NEED_SWAP, fp);
    } else if (wave_info->samp_bit == 32) {
        if (strcaseeq(wave_info->file_desc, WAVE_FORMAT_IEEE_FLOAT_STRING)) {
            len = freadfloattod((double *)data, length, SP_WAV_NEED_SWAP, fp);
        } else {
            len = freadlong32((long *)data, length, SP_WAV_NEED_SWAP, fp);
        }
    } else if (wave_info->samp_bit == 24) {
        len = freadlong24((long *)data, length, SP_WAV_NEED_SWAP, fp);
    } else if (wave_info->samp_bit == 8) {
        if (strcaseeq(wave_info->file_desc, WAVE_FORMAT_ALAW_STRING)) {
            len = freadalaw((short *)data, length, fp);
        } else if (strcaseeq(wave_info->file_desc, WAVE_FORMAT_MULAW_STRING)) {
            len = freadulaw((short *)data, length, fp);
        } else {
            len = freadbyte((short *)data, length, fp);
        }
    } else {
        return -1;
    }
    
    spDebug(80, "spReadWavData", "len = %ld, length = %ld\n", len, length);
    
    return len;
}

long _spWriteWavData(spWaveInfo *wave_info, char *data, long length, FILE *fp)
{
    long len;
    
    if (wave_info == NULL || fp == NULL) return -1;

    if (wave_info->samp_bit == 16) {
        len = fwriteshort((short *)data, length, SP_WAV_NEED_SWAP, fp);
    } else if (wave_info->samp_bit == 64
               && strcaseeq(wave_info->file_desc, WAVE_FORMAT_IEEE_FLOAT_STRING)) {
        len = fwritedouble((double *)data, length, SP_WAV_NEED_SWAP, fp);
    } else if (wave_info->samp_bit == 32) {
        if (strcaseeq(wave_info->file_desc, WAVE_FORMAT_IEEE_FLOAT_STRING)) {
            len = fwritedoubletof((double *)data, length, SP_WAV_NEED_SWAP, fp);
        } else {
            len = fwritelong32((long *)data, length, SP_WAV_NEED_SWAP, fp);
        }
    } else if (wave_info->samp_bit == 24) {
        len = fwritelong24((long *)data, length, SP_WAV_NEED_SWAP, fp);
    } else if (wave_info->samp_bit == 8) {
        if (strcaseeq(wave_info->file_desc, WAVE_FORMAT_ALAW_STRING)) {
            len = fwritealaw((short *)data, length, fp);
        } else if (strcaseeq(wave_info->file_desc, WAVE_FORMAT_MULAW_STRING)) {
            len = fwriteulaw((short *)data, length, fp);
        } else {
            len = fwritebyte((short *)data, length, fp);
        }
    } else {
        return -1;
    }
    
    return len;
}

unsigned long spGetWavSongInfoMask(void)
{
    unsigned long info_mask = SP_SONG_NO_INFO;

    info_mask |= SP_SONG_TITLE_MASK;
    info_mask |= SP_SONG_ARTIST_MASK;
    info_mask |= SP_SONG_COMMENT_MASK;
    info_mask |= SP_SONG_GENRE_MASK;
    info_mask |= SP_SONG_RELEASE_MASK;
    info_mask |= SP_SONG_COPYRIGHT_MASK;
    info_mask |= SP_SONG_ENGINEER_MASK;
    info_mask |= SP_SONG_SOURCE_MASK;
    info_mask |= SP_SONG_SOFTWARE_MASK;
    info_mask |= SP_SONG_SUBJECT_MASK;
    
    return info_mask;
}

long readTextInfo(long info_size, char *info_text,
                  unsigned long info_mask, unsigned long *total_mask, FILE *fp)
{
    long len;
    char buf[1024];
    
    if ((info_size = (long)fread(buf, 1, info_size, fp)) > 0) {
        len = MIN(info_size, SP_SONG_INFO_SIZE - 1);
        strncpy(info_text, buf, len);
        info_text[len] = NUL;

        spConvertKanjiToLocaleCode((unsigned char *)info_text, info_size, SP_KANJI_CODE_SJIS);
    
        *total_mask |= info_mask;
    }
    
    return info_size;
}

spBool spGetWavSongInfo(spSongInfo *song_info, FILE *fp)
{
    spFileOffset offset;
    DWORD list_size = 0;
    DWORD chunk_size = 0;
    DWORD info_size;
    char id[8];
    
    if (song_info == NULL || fp == NULL) return SP_FALSE;

    if ((offset = spTellFile(fp)) <= 0) {
        WAVEHEADER wave_header;
        
        if (readWavHeader(&wave_header, fp, SP_FALSE, SP_FALSE) <= 0
            || spSeekFile(fp, wave_header.nDataBytes, SEEK_CUR) < 0) {
            return SP_FALSE;
        }
    }
    
    for (;;) {
        chunk_size = 0;
        if (fread(id, 1, 4, fp) != 4) {
            spDebug(10, "spGetWavSongInfo", "Can't find LIST chunk.\n");
            return SP_FALSE;
        }

        while (!isalpha(id[0])) {
            spDebug(10, "spGetWavSongInfo", "id[0] is not alphabet.\n");
            id[0] = id[1];
            id[1] = id[2];
            id[2] = id[3];
            if (fread(&id[3], 1, 1, fp) != 1) {
                spDebug(10, "spGetWavSongInfo", "Can't find LIST chunk.\n");
                return SP_FALSE;
            }
        }

        if (spFReadDWORD(&chunk_size, 1, fp) != 1) {
            spDebug(10, "spGetWavSongInfo", "Can't find LIST chunk.\n");
            return SP_FALSE;
        }
        
        spDebug(10, "spGetWavSongInfo", "chunk id: %c%c%c%c\n", id[0], id[1], id[2], id[3]);
        spDebug(10, "spGetWavSongInfo", "chunk_size = %ld\n", chunk_size);
        
        if (strneq("LIST", id, 4)) {
            break;
        }

        spSeekFile(fp, chunk_size, SEEK_CUR);
    }
    
    if (fread(id, 1, 4, fp) != 4 ||
        !strneq("INFO", id, 4)) {
        spDebug(10, "spGetWavSongInfo", "Can't find INFO chunk.\n");
        return SP_FALSE;
    }
    list_size += 4;

    spDebug(10, "spGetWavSongInfo", "read INFO chunk done\n");

    memset((char *)song_info, 0, sizeof(spSongInfo));
    song_info->info_mask = SP_SONG_NO_INFO;
    song_info->track = 1;

    while (list_size < chunk_size) {
        if (fread(id, 1, 4, fp) != 4) {
            break;
        }
        list_size += 4;

        while (!isalpha(id[0])) {
            spDebug(10, "spGetWavSongInfo", "id[0] is not alphabet.\n");
            id[0] = id[1];
            id[1] = id[2];
            id[2] = id[3];
            if (fread(&id[3], 1, 1, fp) != 1) {
                break;
            }
            list_size += 1;
        }

        spDebug(10, "spGetWavSongInfo", "chunk id: %c%c%c%c\n", id[0], id[1], id[2], id[3]);

        info_size = 0;
        spFReadDWORD(&info_size, 1, fp);
        list_size += 4;

        spDebug(10, "spGetWavSongInfo", "info_size = %ld\n", info_size);

        if (strneq("INAM", id, 4)) {
            list_size += readTextInfo(info_size, song_info->title,
                                      SP_SONG_TITLE_MASK, &song_info->info_mask, fp);
        } else if (strneq("IART", id, 4)) {
            list_size += readTextInfo(info_size, song_info->artist,
                                      SP_SONG_ARTIST_MASK, &song_info->info_mask, fp);
        } else if (strneq("ICMT", id, 4)) {
            list_size += readTextInfo(info_size, song_info->comment,
                                      SP_SONG_COMMENT_MASK, &song_info->info_mask, fp);
        } else if (strneq("IGNR", id, 4)) {
            list_size += readTextInfo(info_size, song_info->genre,
                                      SP_SONG_GENRE_MASK, &song_info->info_mask, fp);
        } else if (strneq("ICRD", id, 4)) {
            list_size += readTextInfo(info_size, song_info->release,
                                      SP_SONG_RELEASE_MASK, &song_info->info_mask, fp);
        } else if (strneq("ICOP", id, 4)) {
            list_size += readTextInfo(info_size, song_info->copyright,
                                      SP_SONG_COPYRIGHT_MASK, &song_info->info_mask, fp);
        } else if (strneq("IENG", id, 4)) {
            list_size += readTextInfo(info_size, song_info->engineer,
                                      SP_SONG_ENGINEER_MASK, &song_info->info_mask, fp);
        } else if (strneq("ISRC", id, 4)) {
            list_size += readTextInfo(info_size, song_info->source,
                                      SP_SONG_SOURCE_MASK, &song_info->info_mask, fp);
        } else if (strneq("ISFT", id, 4)) {
            list_size += readTextInfo(info_size, song_info->software,
                                      SP_SONG_SOFTWARE_MASK, &song_info->info_mask, fp);
        } else if (strneq("ISBJ", id, 4)) {
            list_size += readTextInfo(info_size, song_info->subject,
                                      SP_SONG_SUBJECT_MASK, &song_info->info_mask, fp);
        } else {
            spSeekFile(fp, info_size, SEEK_CUR);
        }
    }
    
    spDebug(10, "spGetWavSongInfo", "done\n");
    
    return SP_TRUE;
}

spBool spReadWavSongInfo(spWaveInfo *wave_info, spSongInfo *song_info, FILE *fp)
{
    spLong offset;
    
    if (wave_info == NULL || song_info == NULL || fp == NULL) return SP_FALSE;
    if (wave_info->header_size <= 0) return SP_FALSE;
    
    offset = wave_info->header_size +
        wave_info->length * (spLong)wave_info->num_channel * (spLong)(wave_info->samp_bit / 8);
    if (spSeekFile(fp, offset, SEEK_SET) < 0) {
        return SP_FALSE;
    }

    spDebug(10, "spReadWavSongInfo", "spSeekFile done: %ld\n", offset);

    spGetWavSongInfo(song_info, fp);
    
    return SP_TRUE;
}

static long writeTextInfo(const char *id, const char *text, FILE *fp)
{
    DWORD info_size;
    char buf[SP_SONG_INFO_SIZE];

    strcpy(buf, text);
    spConvertKanjiFromLocaleCode((unsigned char *)buf, SP_SONG_INFO_SIZE, SP_KANJI_CODE_SJIS);
    info_size = (DWORD)strlen(buf) + 1;
    
    if (info_size <= 1) return 0;
    
    fputs(id, fp);
    spFWriteDWORD(&info_size, 1, fp);
    fwrite(buf, 1, info_size, fp);

    return (8 + info_size);
}

spBool spAddWavSongInfo(spSongInfo *song_info, FILE *fp)
{
    spFileOffset offset;
    DWORD chunk_size;

    if (song_info == NULL || fp == NULL 
        || song_info->info_mask == SP_SONG_NO_INFO) return SP_FALSE;
    
    if ((offset = spTellFile(fp)) <= 0) {
        spSeekFile(fp, 0, SEEK_END);
        offset = spTellFile(fp);
    }

    fputs("LIST", fp);
    chunk_size = 4;        /* write dummy size */
    spFWriteDWORD(&chunk_size, 1, fp);
    fputs("INFO", fp);

    if (song_info->info_mask & SP_SONG_TITLE_MASK) {
        chunk_size += writeTextInfo("INAM", song_info->title, fp);
    }
    if (song_info->info_mask & SP_SONG_ARTIST_MASK) {
        chunk_size += writeTextInfo("IART", song_info->artist, fp);
    }
    if (song_info->info_mask & SP_SONG_COMMENT_MASK) {
        chunk_size += writeTextInfo("ICMT", song_info->comment, fp);
    }
    if (song_info->info_mask & SP_SONG_GENRE_MASK) {
        chunk_size += writeTextInfo("IGNR", song_info->genre, fp);
    }
    if (song_info->info_mask & SP_SONG_RELEASE_MASK) {
        chunk_size += writeTextInfo("ICRD", song_info->release, fp);
    }
    if (song_info->info_mask & SP_SONG_COPYRIGHT_MASK) {
        chunk_size += writeTextInfo("ICOP", song_info->copyright, fp);
    }
    if (song_info->info_mask & SP_SONG_ENGINEER_MASK) {
        chunk_size += writeTextInfo("IENG", song_info->engineer, fp);
    }
    if (song_info->info_mask & SP_SONG_SOURCE_MASK) {
        chunk_size += writeTextInfo("ISRC", song_info->source, fp);
    }
    if (song_info->info_mask & SP_SONG_SOFTWARE_MASK) {
        chunk_size += writeTextInfo("ISFT", song_info->software, fp);
    }
    if (song_info->info_mask & SP_SONG_SUBJECT_MASK) {
        chunk_size += writeTextInfo("ISBJ", song_info->subject, fp);
    }
    
    if (chunk_size > 4 && spSeekFile(fp, offset, SEEK_SET) >= 0) {
        /* write real chunk size */
        fputs("LIST", fp);
        spFWriteDWORD(&chunk_size, 1, fp);
    }

    /* update RIFF chunk size */
    spSeekFile(fp, 4, SEEK_SET);
    chunk_size += (DWORD)(offset + 8);
    spFWriteDWORD(&chunk_size, 1, fp);
    
    spSeekFile(fp, 0, SEEK_END);
    
    return SP_TRUE;
}

spBool spWriteWavSongInfo(spWaveInfo *wave_info, spSongInfo *song_info, FILE *fp)
{
    spLong offset;
    
    if (wave_info == NULL || song_info == NULL || fp == NULL) return SP_FALSE;
    if (wave_info->header_size <= 0 || wave_info->length <= 0) return SP_FALSE;

    offset = wave_info->header_size
        + wave_info->length * (spLong)wave_info->num_channel * (spLong)(wave_info->samp_bit / 8);
    if (spSeekFile(fp, offset, SEEK_SET) < 0) {
        return SP_FALSE;
    }

    return spAddWavSongInfo(song_info, fp);
}

#if 1
static long readWaveFormat(WAVEFORMATEX *wave_format, FILE *fp, spBool ex_flag)
{
    long read_size;
    
    if ((read_size = readWaveFormatInfo(&wave_format->wFormatTag, &wave_format->nChannels,
                                        &wave_format->nSamplesPerSec, &wave_format->nAvgBytesPerSec,
                                        &wave_format->nBlockAlign, &wave_format->wBitsPerSample,
                                        fp, ex_flag, SP_FALSE)) <= 0) {
        return 0;
    }

    if (ex_flag == SP_TRUE) {
        if (spFReadWORD(&wave_format->cbSize, 1, fp) <= 0) {
            return 0;
        }
        read_size += 2;
    } else {
        wave_format->cbSize = 0;
    }
    
    spDebug(10, NULL, "%s format, %d channel, %ld samp/sec\n",
            spGetWavFormatLabelFromTag(wave_format->wFormatTag), wave_format->nChannels,
            wave_format->nSamplesPerSec);
    spDebug(10, NULL, "%ld byte/sec, %d block align, %d bits/samp\n",
            wave_format->nAvgBytesPerSec, wave_format->nBlockAlign,
            wave_format->wBitsPerSample);
    spDebug(10, NULL, "read_size = %ld\n", read_size);

    return read_size;
}

static long readBitmapInfoHeader(BITMAPINFOHEADER *bitmap_header, FILE *fp)
{
    long read_size;

    if (spFReadDWORD(&bitmap_header->biSize, 1, fp) <= 0) {
        return 0;
    }
    read_size = 4;

    if (spFReadLONG(&bitmap_header->biWidth, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadLONG(&bitmap_header->biHeight, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadWORD(&bitmap_header->biPlanes, 1, fp) <= 0) {
        return 0;
    }
    read_size += 2;

    if (spFReadWORD(&bitmap_header->biBitCount, 1, fp) <= 0) {
        return 0;
    }
    read_size += 2;

    if (spFReadDWORD(&bitmap_header->biCompression, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadDWORD(&bitmap_header->biSizeImage, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadLONG(&bitmap_header->biXPelsPerMeter, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadLONG(&bitmap_header->biYPelsPerMeter, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadDWORD(&bitmap_header->biClrUsed, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadDWORD(&bitmap_header->biClrImportant, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    spDebug(50, "readBitmapInfoHeader",
            "biSize = %ld, biWidth = %ld, biHeight = %ld, biPlanes = %d, biBitCount = %d, biCompression = %ld, biSizeImage = %ld, biXPelsPerMeter = %ld, biYPelsPerMeter = %ld, biClrUsed = %ld, biClrImportant = %ld\n",
            bitmap_header->biSize, bitmap_header->biWidth, bitmap_header->biHeight,
            bitmap_header->biPlanes, bitmap_header->biBitCount, bitmap_header->biCompression,
            bitmap_header->biSizeImage, bitmap_header->biXPelsPerMeter, bitmap_header->biYPelsPerMeter,
            bitmap_header->biClrUsed, bitmap_header->biClrImportant);
            
    return read_size;
}

static long readSmallRect(SMALL_RECT *rect, FILE *fp)
{
    if (spFReadINT16(&rect->Left, 1, fp) <= 0
        || spFReadINT16(&rect->Top, 1, fp) <= 0
        || spFReadINT16(&rect->Right, 1, fp) <= 0
        || spFReadINT16(&rect->Bottom, 1, fp) <= 0) {
        return 0;
    }

    return 8;
}

static long readDVInfo(DVInfo *dvinfo, FILE *fp)
{
    long read_size;

    if (spFReadDWORD(&dvinfo->dwDVAAuxSrc, 1, fp) <= 0) {
        return 0;
    }
    read_size = 4;

    if (spFReadDWORD(&dvinfo->dwDVAAuxCtl, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;
    
    if (spFReadDWORD(&dvinfo->dwDVAAuxSrc1, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;
    
    if (spFReadDWORD(&dvinfo->dwDVAAuxCtl1, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;
    
    if (spFReadDWORD(&dvinfo->dwDVVAuxSrc, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;

    if (spFReadDWORD(&dvinfo->dwDVVAuxCtl, 1, fp) <= 0) {
        return 0;
    }
    read_size += 4;
    
    spSeekFile(fp, 32, SEEK_CUR);
    read_size += 32;
    
    return read_size;
}

static long readAviHeader(AviMainHeader *avi_header, FILE *fp, spBool after_riff)
{
    long nread;
    DWORD ulSize;
    long head_size = 0;

    if (after_riff == SP_TRUE) {
        head_size = 8;
    } else {
        /* seek start of the file */
        spSeekFile(fp, 0L, SEEK_SET);

        if ((head_size = spReadRiffFourCC(avi_header->RiffID, &ulSize, fp)) <= 0
            || !strneq("RIFF", avi_header->RiffID, 4)) {
            spDebug(1, NULL, "Can't find RIFF chunk.\n");
            return 0;
        }
        avi_header->RiffSize = (DWORD)ulSize;
    }
    
    if (fread(avi_header->AviID, 1, 4, fp) != 4 ||
        !strneq(avi_header->AviID, "AVI ", 4)) {
        spDebug(1, NULL, "Can't find AVI chunk.\n");
        return 0;
    }
    head_size += 4;

    if ((nread = spReadRiffFourCC(avi_header->hdrl.ListID, &ulSize, fp)) <= 0
        || !strneq("LIST", avi_header->hdrl.ListID, 4)) {
        spDebug(1, NULL, "Can't find first LIST chunk.\n");
        return 0;
    }
    avi_header->hdrl.ListSize = (DWORD)ulSize;
    head_size += nread;
    
    if (fread(avi_header->hdrl.fcc, 1, 4, fp) != 4 ||
        !strneq(avi_header->hdrl.fcc, "hdrl", 4)) {
        spDebug(1, NULL, "Can't find header list (hdrl).\n");
        return 0;
    }
    head_size += 4;

    if ((nread = spReadRiffFourCC(avi_header->avihID, &ulSize, fp)) <= 0
        || !strneq("avih", avi_header->avihID, 4)) {
        spDebug(1, NULL, "Can't find 'avih'.\n");
        return 0;
    }
    avi_header->avihSize = (DWORD)ulSize;
    head_size += nread;

    if (spFReadDWORD(&avi_header->dwMicroSecPerFrame, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwMaxBytesPerSec, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwPaddingGranularity, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwFlags, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwTotalFrames, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwInitialFrames, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwStreams, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwSuggestedBufferSize, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwWidth, 1, fp) <= 0
        || spFReadDWORD(&avi_header->dwHeight, 1, fp) <= 0
        || spFReadDWORD(avi_header->dwReserved, 4, fp) <= 0) {
        spDebug(1, NULL, "Can't read header contents.\n");
        return 0;
    }
    head_size += 56;
    if (avi_header->avihSize > 56) {
        spSeekFile(fp, avi_header->avihSize - 56, SEEK_CUR);
    }
        
    spDebug(50, "readAviHeader",
            "dwMicroSecPerFrame = %ld, dwMaxBytesPerSec = %ld, dwPaddingGranularity = %ld, dwFlags = %lx, dwTotalFrames = %ld, dwInitialFrames = %ld, dwStreams = %ld, dwSuggestedBufferSize = %ld, dwWidth = %ld, dwHeight = %ld\n",
            avi_header->dwMicroSecPerFrame,
            avi_header->dwMaxBytesPerSec,
            avi_header->dwPaddingGranularity,
            avi_header->dwFlags,
            avi_header->dwTotalFrames,
            avi_header->dwInitialFrames,
            avi_header->dwStreams,
            avi_header->dwSuggestedBufferSize,
            avi_header->dwWidth,
            avi_header->dwHeight);
    
    spDebug(50, "readAviHeader", "head_size = %ld\n", head_size);
    
    return head_size;
}

long spReadAviHeader(AviMainHeader *avi_header, FILE *fp)
{
    return readAviHeader(avi_header, fp, SP_FALSE);
}

static long readAviStreamHeader(AviMainHeader *avi_header, AviStreamHeader *avi_strh, FILE *fp)
{
    long nread;
    DWORD ulSize;
    long read_size;
    long skip_size;

    if ((read_size = spReadRiffFourCC(avi_strh->strl.ListID, &ulSize, fp)) <= 0
        || !strneq("LIST", avi_strh->strl.ListID, 4)) {
        spDebug(1, NULL, "Can't find 'strl' LIST chunk.\n");
        return 0;
    }
    avi_strh->strl.ListSize = (DWORD)ulSize;
    if (fread(avi_strh->strl.fcc, 1, 4, fp) != 4 ||
        !strneq(avi_strh->strl.fcc, "strl", 4)) {
        spDebug(1, NULL, "Can't find 'strl'.\n");
        return 0;
    }
    read_size += 4;

    if ((nread = spReadRiffFourCC(avi_strh->strhID, &ulSize, fp)) <= 0
        || !strneq("strh", avi_strh->strhID, 4)) {
        spDebug(1, NULL, "Can't find 'strh' chunk.\n");
        return 0;
    }
    avi_strh->strhSize = (DWORD)ulSize;
    read_size += nread;
    
    if (fread(avi_strh->fccType, 1, 4, fp) != 4) {
        spDebug(1, NULL, "Can't find source type.\n");
        return 0;
    }
    read_size += 4;
    spDebug(100, "readAviStreamHeader", "avi_strh->fccType = %c%c%c%c\n",
            avi_strh->fccType[0], avi_strh->fccType[1], avi_strh->fccType[2], avi_strh->fccType[3]);
    
    if (fread(avi_strh->fccHandler, 1, 4, fp) != 4) {
        spDebug(1, NULL, "Can't find codec type.\n");
        return 0;
    }
    read_size += 4;
    spDebug(100, "readAviStreamHeader", "avi_strh->fccHandler = %c%c%c%c\n",
            avi_strh->fccHandler[0], avi_strh->fccHandler[1], avi_strh->fccHandler[2], avi_strh->fccHandler[3]);
    
    if (spFReadDWORD(&avi_strh->dwFlags, 1, fp) <= 0
        || spFReadWORD(&avi_strh->wPriority, 1, fp) <= 0
        || spFReadWORD(&avi_strh->wLanguage, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwInitialFrames, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwScale, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwRate, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwStart, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwLength, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwSuggestedBufferSize, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwQuality, 1, fp) <= 0
        || spFReadDWORD(&avi_strh->dwSampleSize, 1, fp) <= 0) {
        spDebug(1, NULL, "Can't read header contents.\n");
        return 0;
    }
    read_size += 40;                /* DWORD*9 + WORD*2 = 40 */

    spDebug(50, "readAviStreamHeader",
            "dwFlags = %lx, wPriority = %d, wLanguage = %d, dwInitalFrames = %ld, dwScale = %ld, dwRate = %ld, dwStart = %ld, dwLength = %ld, dwSuggestedBufferSize = %ld, dwQuality = %ld, dwSampleSize = %ld\n",
            avi_strh->dwFlags, avi_strh->wPriority, avi_strh->wLanguage, avi_strh->dwInitialFrames,
            avi_strh->dwScale, avi_strh->dwRate, avi_strh->dwStart, avi_strh->dwLength,
            avi_strh->dwSuggestedBufferSize, avi_strh->dwQuality, avi_strh->dwSampleSize);
    
    if ((nread = readSmallRect(&avi_strh->rcFrame, fp)) <= 0) {
        spDebug(1, NULL, "Can't find rcFrame.\n");
        return 0;
    }
    read_size += nread;

    skip_size = avi_strh->strhSize - (read_size - 20);
    if (skip_size > 0) {
        spSeekFile(fp, skip_size, SEEK_CUR);
        read_size += skip_size;
    }
    
    if ((nread = spReadRiffFourCC(avi_strh->strfID, &ulSize, fp)) <= 0
        || !strneq("strf", avi_strh->strfID, 4)) {
        spDebug(1, NULL, "Can't find 'strf' chunk.\n");
        return 0;
    }
    avi_strh->strfSize = (DWORD)ulSize;
    read_size += nread;
    
    spDebug(50, "readAviStreamHeader", "read_size = %ld\n", read_size);
    
    return read_size;
}

spBool _spFreeAviIndexChunk(AviIndexChunk *index_chunk)
{
    if (index_chunk != NULL && index_chunk->aIndex != NULL) {
        xfree(index_chunk->aIndex);
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

long spReadAviIndexChunk(AviMainHeader *avi_header, long remain_size,
                         AviIndexChunk *index_chunk, FILE *fp)
{
    DWORD k;
    DWORD dwSize;
    long nread;
    long read_size;
    long stride;
    long skip_size;
    DWORD ulSize;

    if (remain_size <= 8) {
        spDebug(50, "readAviIndexChunk", "only strh and strf are included\n");
        return 0;
    }

    read_size = 0;

    do {
        if ((nread = spReadRiffFourCC(index_chunk->header.fcc, &ulSize, fp)) <= 0) {
            return 0;
        }
        index_chunk->header.dwSize = ulSize;
        read_size += nread;

        if (strneq("indx", index_chunk->header.fcc, 4) || strneq("ix", index_chunk->header.fcc, 2)) {
            spDebug(100, "readAviIndexChunk", "indx found\n");
            break;
        }
        spDebug(100, "readAviIndexChunk", "**** indx NOT found ****, remain_size = %ld\n", remain_size);

        dwSize = ((index_chunk->header.dwSize + 1) / 2) * 2; /* alignment in WORD */
        spSeekFile(fp, dwSize, SEEK_CUR);
        remain_size -= (nread + dwSize);
        read_size += dwSize;

        memset(index_chunk, 0, 8);
        
        if (remain_size <= 0) {
            return read_size;
        }
    } while (1);
    spDebug(100, "readAviIndexChunk", "index_chunk->header.fcc = %c%c%c%c, size = %ld\n",
            index_chunk->header.fcc[0], index_chunk->header.fcc[1], index_chunk->header.fcc[2], index_chunk->header.fcc[3], index_chunk->header.dwSize);

    if (spFReadWORD(&index_chunk->header.wLongsPerEntry, 1, fp) <= 0
        || fread(&index_chunk->header.bIndexSubType, 1, 1, fp) <= 0
        || fread(&index_chunk->header.bIndexType, 1, 1, fp) <= 0
        || spFReadDWORD(&index_chunk->header.nEntriesInUse, 1, fp) <= 0
        || fread(index_chunk->header.chunkID, 1, 4, fp) != 4
        || fread(index_chunk->header.qwOffset, 1, 8, fp) != 8
        || spFReadDWORD(&index_chunk->header.dwReserved, 1, fp) <= 0) {
        return 0;
    }
    read_size += 24;
        
    spDebug(100, "readAviIndexChunk", "wLongsPerEntry = %d, bIndexSubType = %d, bIndexType = %d, nEntriesInUse = %ld, chunkID = %c%c%c%c\n",
            index_chunk->header.wLongsPerEntry, index_chunk->header.bIndexSubType,
            index_chunk->header.bIndexType, index_chunk->header.nEntriesInUse,
            index_chunk->header.chunkID[0], index_chunk->header.chunkID[1], index_chunk->header.chunkID[2], index_chunk->header.chunkID[3]);

    stride = 4 * index_chunk->header.wLongsPerEntry;
    index_chunk->aIndex = xalloc(stride * index_chunk->header.nEntriesInUse, char);
    
    if (index_chunk->header.bIndexType == AVI_INDEX_OF_CHUNKS) {
        AviStdIndexEntry *std_index;
        
        skip_size = stride - 8;
        std_index = (AviStdIndexEntry *)index_chunk->aIndex;
        
        for (k = 0; k < index_chunk->header.nEntriesInUse; k++) {
            if (spFReadDWORD(&std_index[k].dwOffset, 1, fp) <= 0
                || spFReadDWORD(&std_index[k].dwSize, 1, fp) <= 0) {
                return 0;
            }
            spDebug(100, "readAviIndexChunk", "std_index[%ld]: dwOffset = %ld, dwSize = %ld\n",
                    k, std_index[k].dwOffset, 0x7fffffff & std_index[k].dwSize);
            
            if (skip_size > 0) {
                spSeekFile(fp, skip_size, SEEK_CUR);
            }
            
            read_size += stride;
        }
    } else if (index_chunk->header.bIndexType == AVI_INDEX_OF_INDEXES) {
        AviSuperIndexEntry *super_index;
        
        skip_size = stride - 16;
        super_index = (AviSuperIndexEntry *)index_chunk->aIndex;
        spDebug(100, "readAviIndexChunk", "stride = %ld, skip_size = %ld\n", stride, skip_size);
        
        for (k = 0; k < index_chunk->header.nEntriesInUse; k++) {
            spDebug(100, "readAviIndexChunk", "k = %ld / %ld\n", k, index_chunk->header.nEntriesInUse);
            if (fread(super_index[k].qwOffset, 1, 8, fp) != 8
                || spFReadDWORD(&super_index[k].dwSize, 1, fp) <= 0
                || spFReadDWORD(&super_index[k].dwDuration, 1, fp) <= 0) {
                return 0;
            }
            spDebug(100, "readAviIndexChunk", "super_index[%ld]: dwSize = %ld, dwDuration = %ld\n",
                    k, super_index[k].dwSize, super_index[k].dwDuration);
            
            if (skip_size > 0) {
                spSeekFile(fp, skip_size, SEEK_CUR);
            }
            
            read_size += stride;
        }
    }
    
    return read_size;
}

spBool spIsAviStreamValid(AviStreamHeader *avi_strh)
{
    if (avi_strh != NULL && avi_strh->strl.ListSize >= 68) {
        spDebug(100, "spIsAviStreamValid", "strl.LtrfSize = %ld, strhSize = %ld\n",
                avi_strh->strl.ListSize/* >= 68 */, avi_strh->strhSize/* >= 56*/);
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spIsAviStreamVideo(AviStreamHeader *avi_strh)
{
    if (strneq("vids", avi_strh->fccType, 4)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spIsAviStreamIAVS(AviStreamHeader *avi_strh)
{
    if (strneq("iavs", avi_strh->fccType, 4)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spIsAviStreamAudio(AviStreamHeader *avi_strh)
{
    if (strneq("auds", avi_strh->fccType, 4)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spIsAviStreamSubtitle(AviStreamHeader *avi_strh)
{
    if (strneq("txts", avi_strh->fccType, 4)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

long spReadAviStreamAudioFormat(AviStreamHeader *avi_strh, WAVEFORMATEX *format, FILE *fp)
{
    long read_size;
    long remain_size;
    spBool ex_flag;

    if (avi_strh->strfSize >= 20) {
        ex_flag = SP_TRUE;
    } else if (avi_strh->strfSize >= 14) {
        ex_flag = SP_FALSE;
    } else {
        return 0;
    }
        
    read_size = readWaveFormat(format, fp, ex_flag);
    spDebug(50, "spReadAviStreamAudioFormat", "read_size = %ld, strfSize = %ld\n",
            read_size, avi_strh->strfSize);
    if (read_size <= 0 || (DWORD)read_size > avi_strh->strfSize) {
        return 0;
    }

    remain_size = (long)(avi_strh->strfSize - (DWORD)read_size);
    if (remain_size > 0) {
        spSeekFile(fp, remain_size, SEEK_CUR);
        read_size += remain_size;
    }

    return read_size;
}

long spReadAviStreamVideoFormat(AviStreamHeader *avi_strh, BITMAPINFOHEADER *format, FILE *fp)
{
    long read_size;
    long remain_size;

    read_size = readBitmapInfoHeader(format, fp);
    spDebug(50, "spReadAviStreamVideoFormat", "read_size = %ld, strfSize = %ld\n",
            read_size, avi_strh->strfSize);
    if (read_size <= 0 || (DWORD)read_size > avi_strh->strfSize) {
        return 0;
    }

    remain_size = (long)(avi_strh->strfSize - (DWORD)read_size);
    if (remain_size > 0) {
        spSeekFile(fp, remain_size, SEEK_CUR);
        read_size += remain_size;
    }
    
    return read_size;
}

long spReadAviStreamDVInfo(AviStreamHeader *avi_strh, DVInfo *dvinfo, FILE *fp)
{
    long read_size;

    read_size = readDVInfo(dvinfo, fp);
    spDebug(50, "spReadAviStreamVideoFormat", "read_size = %ld, strfSize = %ld\n",
            read_size, avi_strh->strfSize);
    if (read_size <= 0 || (DWORD)read_size > avi_strh->strfSize) {
        return 0;
    }

    return read_size;
}

static long readAviOldIndexEntry(AviOldIndexEntry *entry, FILE *fp, spBool id_only)
{
    long nread;
    DWORD ulSize;
    long read_size;
    
    if ((nread = spReadRiffFourCC(entry->chunkID, &ulSize, fp)) <= 0) {
        spDebug(50, "readAviOldIndexEntry", "Can't read chunkID.\n");
        return 0;
    }
    entry->dwFlags = (DWORD)ulSize;
    read_size = nread;
 
    if (id_only == SP_TRUE) {
        spSeekFile(fp, 8, SEEK_CUR);
    } else {
        if (spFReadDWORD(&entry->dwChunkOffset, 1, fp) <= 0
            || spFReadDWORD(&entry->dwChunkLength, 1, fp) <= 0) {
            return 0;
        }
    }
    read_size += 8;

   spDebug(50, "readAviOldIndexEntry",
           "id = %c%c%c%c, dwChunkOffset = %ld, dwChunkLength = %ld, read_size = %ld\n",
            entry->chunkID[0], entry->chunkID[1], entry->chunkID[2], entry->chunkID[3],
           entry->dwChunkOffset, entry->dwChunkLength, read_size);
   
    return read_size;
}

static void resetAviStdIndexChunk(AviStdIndexChunk *std_index_chunk, long count)
{
    memset(std_index_chunk, 0, sizeof(AviStdIndexChunk));
    memcpy(std_index_chunk->header.fcc, "indx", 4);
    std_index_chunk->header.wLongsPerEntry = 2;
    std_index_chunk->header.bIndexType = AVI_INDEX_OF_CHUNKS;
    std_index_chunk->header.nEntriesInUse = count;
    std_index_chunk->aIndex = xalloc(count, AviStdIndexEntry);
    std_index_chunk->header.dwSize = 24 + 8 * count;

    return;
}

static void convAviIndexEntryOldToStd(AviOldIndexEntry *old_index_entry, /* (i) */
                                      AviStdIndexEntry *std_index_entry) /* (o) */
{
    std_index_entry->dwOffset = old_index_entry->dwChunkOffset + 8;
    std_index_entry->dwSize = old_index_entry->dwChunkLength;
    spDebug(50, "convAviIndexEntryOldToStd",
            "update: %c%c%c%c: dwOffset = %ld, dwSize = %ld\n",
            old_index_entry->chunkID[0], old_index_entry->chunkID[1],
            old_index_entry->chunkID[2], old_index_entry->chunkID[3],
            std_index_entry->dwOffset, std_index_entry->dwSize);
                        
    if (!(old_index_entry->dwFlags & AVIIF_KEYFRAME)) {
        std_index_entry->dwSize |= 0x80000000;
    }

    return;
}

static long convAviOldIndexToStdIndexList(AviMainHeader *avi_header,           /* (i) */
                                          AviOldIndexChunk *old_index_chunk,   /* (i) */
                                          DWORD moviPos,                       /* (i) */
                                          AviStreamInfo *stream_infos)         /* (o) */
{
    DWORD k, l;
    long nread;
    long remain_size;
    long *counts;
    long total_count = 0;
    char numstr[3];
    long stream_id;
    spBool from_movi = SP_TRUE;

    counts = xalloc(avi_header->dwStreams, long);
    
    nread = 16;
    
    if (old_index_chunk->aIndex[0].dwChunkOffset > 20) {
        from_movi = SP_FALSE;
    }
    
    numstr[2] = '\0';
    
    for (k = 0;; k++) {
        memset(counts, 0, sizeof(long) * avi_header->dwStreams);
        remain_size = old_index_chunk->idx1Size;
        
        for (l = 0; remain_size > 0; l++) {
            remain_size -= nread;

            memcpy(numstr, old_index_chunk->aIndex[l].chunkID, 2);
            stream_id = atoi(numstr);
            
            if (k == 1) {
                if (counts[stream_id] == 0) {
                    memcpy(stream_infos[stream_id].std_index_list[0].header.chunkID,
                           old_index_chunk->aIndex[l].chunkID, 4);
                }
                convAviIndexEntryOldToStd(&old_index_chunk->aIndex[l],
                                          &stream_infos[stream_id].std_index_list[0].aIndex[counts[stream_id]]);
            }
            
            ++counts[stream_id];
        }

        if (k == 1) {
            break;
        }

        total_count = 0;
        for (l = 0; l < avi_header->dwStreams; l++) {
            stream_infos[l].num_std_index_list = 1;
            stream_infos[l].std_index_list = xalloc(1, AviStdIndexChunk);
            
            resetAviStdIndexChunk(&stream_infos[l].std_index_list[0], counts[l]);
            if (from_movi == SP_TRUE) spAddC64(stream_infos[l].std_index_list[0].header.qwOffset,
                                               moviPos, SP_WAV_NEED_SWAP);

            total_count += counts[l];
        }
    }

    spDebug(50, "convAviOldIndexToStdIndexList", "total_count = %ld\n", total_count);

    xfree(counts);
    
    return total_count;
}

static long readAviOldIndexChunk(AviMainHeader *avi_header,         /* (i) */
                                 AviOldIndexChunk *old_index_chunk, /* (i) */
                                 AviStreamInfo *stream_infos,       /* (o) */
                                 FILE *fp)                          /* (i) */
{
    long k;
    char idx1ID[4];
    DWORD idx1Size;
    DWORD moviPos = 0;
    long nread;
    long read_size = 0;
    long count;
        
    spDebug(50, "readAviOldIndexChunk", "AVI 1.0 (no indx chunk)\n");

    read_size = 20 + avi_header->hdrl.ListSize;
    spSeekFile(fp, read_size, SEEK_SET);
    do {
        if ((nread = spReadRiffFourCC(idx1ID, &idx1Size, fp)) <= 0) {
            return 0;
        }
        read_size += nread;
        spDebug(10, "readAviOldIndexChunk", "id = %c%c%c%c, size = %ld, read_size = %ld\n",
                idx1ID[0], idx1ID[1], idx1ID[2], idx1ID[3], idx1Size, read_size);
            
        if (strneq("LIST", idx1ID, 4)) {
            if (fread(idx1ID, 1, 4, fp) != 4) {
                return 0;
            }
            if (strneq("movi", idx1ID, 4)) {
                moviPos = read_size;
            }
            read_size += 4;
            idx1Size -= 4;
        } else if (strneq("idx1", idx1ID, 4)) {
            break;
        }
            
        spSeekFile(fp, idx1Size, SEEK_CUR);
        read_size += idx1Size;
    } while (1);

    read_size = 0;
    count = idx1Size / 16;
    memcpy(old_index_chunk->idx1ID, "idx1", 4);
    old_index_chunk->idx1Size = (DWORD)idx1Size;
    old_index_chunk->aIndex = xalloc(count, AviOldIndexEntry);
    spDebug(50, "readAviOldIndexChunk", "moviPos = %ld, idx1Size = %ld, count = %ld\n",
            moviPos, idx1Size, count);
    
    for (k = 0; k < count; k++) {
        if ((nread = readAviOldIndexEntry(&old_index_chunk->aIndex[k], fp, SP_FALSE)) <= 0) {
            return 0;
        }
        read_size += nread;
    }

    return convAviOldIndexToStdIndexList(avi_header, old_index_chunk, moviPos, stream_infos);
}

static DWORD getSuperIndexNumEntries(AviSuperIndexChunk *super_index)
{
    if (super_index->header.bIndexType == AVI_INDEX_OF_CHUNKS) {
        return 1L;
    } else {
        spDebug(50, "getSuperIndexNumEntries", "super_index->header.nEntriesInUse = %ld\n",
                super_index->header.nEntriesInUse);
        
        return super_index->header.nEntriesInUse;
    }
}

static long readAviStdIndexChunk(AviMainHeader *avi_header,             /* (i) */
                                 AviSuperIndexChunk *super_index_chunk, /* (i) */
                                 DWORD number,                          /* (i) */
                                 AviStdIndexChunk *std_index_chunk,     /* (o) */
                                 FILE *fp)                              /* (i) */
{
    long read_size = 0;

    spDebug(50, "readAviStdIndexChunk", "super_index_chunk id = %c%c%c%c, dwSize = %ld\n",
            super_index_chunk->header.chunkID[0],
            super_index_chunk->header.chunkID[1],
            super_index_chunk->header.chunkID[2],
            super_index_chunk->header.chunkID[3],
            super_index_chunk->header.dwSize);
    
    memset(std_index_chunk, 0, sizeof(AviStdIndexChunk));
    
    if (super_index_chunk->header.bIndexType == AVI_INDEX_OF_CHUNKS) {
        /* case where super_index_chunk is AviStdIndexChunk */
        long stride;
        
        memcpy(std_index_chunk, super_index_chunk, sizeof(AviStdIndexChunk));

        /* copy aIndex contents */
        stride = 4 * super_index_chunk->header.wLongsPerEntry;
        read_size = stride * super_index_chunk->header.nEntriesInUse;
        std_index_chunk->aIndex = (AviStdIndexEntry *)xalloc(read_size, char);
        memcpy(std_index_chunk->aIndex, super_index_chunk->aIndex, read_size);
    } else {
        spSeekFileC64(fp, super_index_chunk->aIndex[number].qwOffset,
                      SP_WAV_NEED_SWAP, SEEK_SET);
        read_size = spReadAviIndexChunk(avi_header, super_index_chunk->aIndex[number].dwSize,
                                        (AviIndexChunk *)std_index_chunk, fp);
    }

    return read_size;
}

spBool spFreeAviStreamInfo(AviStreamInfo *stream_info)
{
    DWORD k;
    
    if (stream_info != NULL) {
        /*for (k = 0; k < MAX(stream_info->num_std_index_list, 1); k++) {*/
        for (k = 0; k < stream_info->num_std_index_list; k++) {
            spFreeAviIndexChunk(&stream_info->std_index_list[k]);
        }
        spFreeAviIndexChunk(&stream_info->super_index);

        if (stream_info->strn != NULL) xfree(stream_info->strn);

        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spFreeAviStreamInfos(AviMainHeader *avi_header,   /* (i) */
                            AviStreamInfo *stream_infos) /* (t) */
{
    DWORD k;
    
    if (stream_infos != NULL) {
        for (k = 0; k < avi_header->dwStreams; k++) {
            spFreeAviStreamInfo(&stream_infos[k]);
        }
        xfree(stream_infos);
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

AviStreamInfo *spReadAviStreamInfos(                               /* (o) */
                                    AviMainHeader *avi_header, /* (i) */
                                    FILE *fp)                       /* (i) */
{
    DWORD k;
    long nread;
    long strh_whole_size;
    long strl_whole_size;
    long total_read_size;
    long remain_size;
    AviStreamHeader *pstrh;
    char stream_str[4];
    AviStreamInfo *stream_infos;

    remain_size = 0;
    total_read_size = 0;

    stream_infos = xalloc(avi_header->dwStreams, AviStreamInfo);
    memset(stream_infos, 0, sizeof(AviStreamInfo) * avi_header->dwStreams);
    
    for (k = 0; k < avi_header->dwStreams; k++) {
        pstrh = &stream_infos[k].strh;
        if ((strh_whole_size = readAviStreamHeader(avi_header, pstrh, fp)) <= 0) {
            return 0;
        }
        total_read_size += strh_whole_size;

        sprintf(stream_str, "%02ld", (long)k);
        
        strl_whole_size = 0;
        if (spIsAviStreamVideo(pstrh) == SP_TRUE) {
            if ((strl_whole_size = spReadAviStreamVideoFormat(pstrh,
                                                              &stream_infos[k].format.video, fp)) <= 0) {
                return 0;
            }
            memcpy(stream_infos[k].super_index.header.chunkID, stream_str, 2);
            memcpy(stream_infos[k].super_index.header.chunkID + 2, "db", 2);
        } else if (spIsAviStreamAudio(pstrh) == SP_TRUE) {
            if ((strl_whole_size = spReadAviStreamAudioFormat(pstrh,
                                                              &stream_infos[k].format.audio, fp)) <= 0) {
                return 0;
            }
            memcpy(stream_infos[k].super_index.header.chunkID, stream_str, 2);
            memcpy(stream_infos[k].super_index.header.chunkID + 2, "wb", 2);
        }
        total_read_size += strl_whole_size;
        
        remain_size = pstrh->strl.ListSize - (strh_whole_size - 8) - strl_whole_size;
        spDebug(50, "spReadAviAVStreamsInfo",
                "strh_whole_size = %ld, strl_whole_size = %ld, total_read_size = %ld, remain_size = %ld\n",
                strh_whole_size, strl_whole_size, total_read_size, remain_size);
        if (remain_size > 0) {
            AviIndexChunk *index_chunk = NULL;

            index_chunk = (AviIndexChunk *)&stream_infos[k].super_index;

            if ((nread = spReadAviIndexChunk(avi_header, remain_size, index_chunk, fp)) > 0) {
                remain_size -= nread;
            }
            
            spSeekFile(fp, remain_size, SEEK_CUR);
            total_read_size += remain_size;
        }
    }

    spDebug(50, "spReadAviAVStreamsInfo", "done: total_read_size = %ld, remain_size = %ld\n",
            total_read_size, remain_size);
    
    return stream_infos;
}

static AviStdIndexChunk *readAviStdIndexList(                                 /* (o) */
                                             AviMainHeader *avi_header,       /* (i) */
                                             AviSuperIndexChunk *super_index, /* (i) */
                                             DWORD *plist_count,              /* (o) */
                                             FILE *fp)                        /* (i) */
{
    DWORD k;
    long nread;
    long read_size = 0;
    DWORD list_count;
    AviStdIndexChunk *std_index_list;

    list_count = getSuperIndexNumEntries(super_index);
    *plist_count = list_count;
    
    if (list_count <= 0) return NULL;
    
    /*std_index_list = xalloc(MAX(list_count, 1), AviStdIndexChunk);*/
    std_index_list = xalloc(list_count, AviStdIndexChunk);
    memset(&std_index_list[0], 0, sizeof(AviStdIndexChunk));
        
    for (k = 0; k < list_count; k++) {
        if ((nread = readAviStdIndexChunk(avi_header, super_index, k,
                                          &std_index_list[k], fp)) <= 0) {
            xfree(std_index_list);
            *plist_count = 0;
            return NULL;
        }
        read_size += nread;
    }

    spDebug(50, "readAviStdIndexList", "read_size = %ld\n", read_size);
    
    return std_index_list;
}

#if defined(__STDC__) || defined(_WIN32) || defined(MACOS)
static int qsort_cmp(const void *x, const void *y)
#else
static int qsort_cmp(char *x, char *y)
#endif
{
    AviOldIndexEntry *tx, *ty;

    tx = (AviOldIndexEntry *)x;
    ty = (AviOldIndexEntry *)y;

    if (tx->dwChunkOffset < ty->dwChunkOffset) return -1;
    if (tx->dwChunkOffset > ty->dwChunkOffset) return 1;
    return 0;
}

static long convAviStdIndexListToOldIndex(AviMainHeader *avi_header,         /* (i) */
                                          AviStreamInfo *stream_infos,       /* (i) */
                                          DWORD total_count,                 /* (i) */
                                          AviOldIndexChunk *old_index_chunk) /* (o) */
{
    DWORD k, l, m;
    DWORD count;
    spULong base_offset;
    AviStdIndexChunk *std_index_chunk;
    AviStdIndexEntry *entry;

    memcpy(old_index_chunk->idx1ID, "idx1", 4);
    old_index_chunk->idx1Size = total_count * 16;
    old_index_chunk->aIndex = xalloc(total_count, AviOldIndexEntry);
    
    count = 0;
    for (k = 0; k < avi_header->dwStreams; k++) {
        for (l = 0; l < stream_infos[k].num_std_index_list; l++) {
            std_index_chunk = &stream_infos[k].std_index_list[l];
            base_offset = spC64ToULong(std_index_chunk->header.qwOffset, SP_WAV_NEED_SWAP);
            
            for (m = 0; m < std_index_chunk->header.nEntriesInUse; m++) {
                entry = &std_index_chunk->aIndex[m];
            
                memcpy(old_index_chunk->aIndex[count].chunkID, std_index_chunk->header.chunkID, 4);
                old_index_chunk->aIndex[count].dwChunkOffset = entry->dwOffset + (DWORD)base_offset;
                old_index_chunk->aIndex[count].dwChunkLength = 0x7fffffff & entry->dwSize;
                if ((entry->dwSize & 0x80000000) == 0) {
                    old_index_chunk->aIndex[count].dwFlags |= AVIIF_KEYFRAME;
                }
                spDebug(50, "convAviStdIndexListToOldIndex",
                        "before sort: aIndex[%ld].dwChunkOffset = %ld\n", count, old_index_chunk->aIndex[count].dwChunkOffset);

                ++count;
            }
        }
    }

    /* sort by dwChunkOffset */
    qsort(old_index_chunk->aIndex, count, sizeof(AviOldIndexEntry), qsort_cmp);
#if 0
    {
        for (k = 0; k < count; k++) {
            spDebug(50, "convAviStdIndexListToOldIndex",
                    "after sort: aIndex[%ld].dwChunkOffset = %ld\n", k, old_index_chunk->aIndex[k].dwChunkOffset);
        }
    }
#endif
    
    spDebug(50, "convAviStdIndexListToOldIndex", "total_count = %ld, count = %ld\n",
            total_count, count);
    
    return count;
}

DWORD spReadAviStdIndex(AviMainHeader *avi_header,   /* (i) */
                        AviStreamInfo *stream_infos, /* (i/o) */
                        AviOldIndexChunk *old_index, /* (o) */
                        FILE *fp)                    /* (i) */
{
    DWORD k, l;
    DWORD total_count = 0L;

    for (k = 0; k < avi_header->dwStreams; k++) {
        stream_infos[k].std_index_list = readAviStdIndexList(avi_header, &stream_infos[k].super_index,
                                                             &stream_infos[k].num_std_index_list, fp);
        for (l = 0; l < stream_infos[k].num_std_index_list; l++) {
            total_count += stream_infos[k].std_index_list[l].header.nEntriesInUse;
        }
    }

    if (total_count == 0L) {
        /* case where file format is AVI 1.0 */
        total_count = readAviOldIndexChunk(avi_header, old_index, stream_infos, fp);
    } else {
        convAviStdIndexListToOldIndex(avi_header, stream_infos, total_count, old_index);
    }

    return total_count;
}

int spSeekAviChunkByStdIndex(AviMainHeader *avi_header,         /* (i) */
                             AviStdIndexChunk *std_index_chunk, /* (i) */
                             DWORD count,                        /* (i) */
                             FILE *fp)                          /* (i) */
{
    int status;
    
    if ((status = spSeekFileC64(fp, std_index_chunk->header.qwOffset,
                                SP_WAV_NEED_SWAP, SEEK_SET)) != 0) {
        return status;
    }

    return spSeekFile(fp, std_index_chunk->aIndex[count].dwOffset - 8, SEEK_CUR);
}
#endif

